import configparser
import logging
import multiprocessing as mp
import pickle
import os
from abc import ABC, abstractmethod
from time import sleep
from typing import Dict

import numpy as np

from autoconf import conf
from autofit.mapper import model_mapper as mm
from autofit.non_linear.initializer import Initializer
from autofit.non_linear.paths import Paths, convert_paths
from autofit.text import formatter
from autofit.text import model_text
from autofit.text import samples_text
from autofit import exc

logging.basicConfig()
logger = logging.getLogger(__name__)  # TODO: Logging issue


class NonLinearSearch(ABC):
    @convert_paths
    def __init__(
        self,
        paths=None,
        initializer=None,
        iterations_per_update=None,
        number_of_cores=1,
    ):
        """Abstract base class for non-linear searches.

        This class sets up the file structure for the non-linear search, which are standardized across all non-linear
        searches.

        Parameters
        ------------
        paths : af.Paths
            A class that manages all paths, e.g. where the phase outputs are stored, the non-linear search samples,
            backups, etc.
        sigma : float
            The error-bound value that linked Gaussian prior withs are computed using. For example, if sigma=3.0,
            parameters will use Gaussian Priors with widths coresponding to errors estimated at 3 sigma confidence.
        initializer : non_linear.initializer.Initializer
            Generates the initialize samples of non-linear parameter space (see autofit.non_linear.initializer).
        """

        if paths.non_linear_name is "":
            paths.non_linear_name = self.config("tag", "name")

        if paths.non_linear_tag is "":
            paths.non_linear_tag_function = lambda: self.tag

        self.paths = paths

        self.skip_completed = conf.instance.general.get(
            "output", "skip_completed", bool
        )
        self.log_file = conf.instance.general.get("output", "log_file", str).replace(
            " ", ""
        )

        if initializer is None:
            self.initializer = Initializer.from_config(config=self.config)
        else:
            self.initializer = initializer

        self.iterations_per_update = (
            self.config("updates", "iterations_per_update", int)
            if iterations_per_update is None
            else iterations_per_update
        )

        self.log_every_update = self.config("updates", "log_every_update", int)
        self.backup_every_update = self.config("updates", "backup_every_update", int)
        self.visualize_every_update = self.config(
            "updates", "visualize_every_update", int
        )
        self.model_results_every_update = self.config(
            "updates", "model_results_every_update", int
        )

        self.should_log = IntervalCounter(self.log_every_update)
        self.should_backup = IntervalCounter(self.backup_every_update)
        self.should_visualize = IntervalCounter(self.visualize_every_update)
        self.should_output_model_results = IntervalCounter(
            self.model_results_every_update
        )

        self.silence = self.config("printing", "silence", bool)

        self.number_of_cores = number_of_cores

        self._in_phase = False

    class Fitness:
        def __init__(self, paths, model, analysis, samples_from_model, pool_ids=None):

            self.paths = paths
            self.max_log_likelihood = -np.inf
            self.analysis = analysis

            self.model = model
            self.samples_from_model = samples_from_model

            self.pool_ids = pool_ids

        def fit_instance(self, instance):

            log_likelihood = self.analysis.log_likelihood_function(instance=instance)

            if self.analysis.log_likelihood_cap is not None:
                if log_likelihood > self.analysis.log_likelihood_cap:
                    log_likelihood = self.analysis.log_likelihood_cap

            if log_likelihood > self.max_log_likelihood:

                if self.pool_ids is not None:
                    if mp.current_process().pid != min(self.pool_ids):
                        return log_likelihood

                self.max_log_likelihood = log_likelihood

            return log_likelihood

        def log_likelihood_from_parameters(self, parameters):
            instance = self.model.instance_from_vector(vector=parameters)
            log_likelihood = self.fit_instance(instance)
            return log_likelihood

        def log_posterior_from_parameters(self, parameters):
            log_likelihood = self.log_likelihood_from_parameters(parameters=parameters)
            log_priors = self.model.log_priors_from_vector(vector=parameters)
            return log_likelihood + sum(log_priors)

        def figure_of_merit_from_parameters(self, parameters):
            """The figure of merit is the value that the non-linear search uses to sample parameter space. This varies
            between different non-linear search algorithms, for example:

                - The *Optimizer* *PySwarms* uses the chi-squared value, which is the -2.0*log_posterior.
                - The *MCMC* algorithm *Emcee* uses the log posterior.
                - Nested samplers such as *Dynesty* use the log likelihood.
            """
            raise NotImplementedError()

        @staticmethod
        def prior(cube, model):

            # NEVER EVER REFACTOR THIS LINE! Haha.

            phys_cube = model.vector_from_unit_vector(unit_vector=cube)

            for i in range(len(phys_cube)):
                cube[i] = phys_cube[i]

            return cube

        @staticmethod
        def fitness(cube, model, fitness_function):
            return fitness_function(instance=model.instance_from_vector(cube))

        @property
        def samples(self):
            return self.samples_from_model(model=self.model)

        @property
        def resample_figure_of_merit(self):
            """If a sample raises a FitException, this value is returned to signify that the point requires resampling or
             should be given a likelihood so low that it is discard."""
            return -np.inf

    def fit(self, model, analysis: "Analysis", info=None) -> "Result":
        """ Fit a model, M with some function f that takes instances of the
        class represented by model M and gives a score for their fitness.

        A model which represents possible instances with some dimensionality is fit.

        The analysis provides two functions. One visualises an instance of a model and the
        other scores an instance based on how well it fits some data. The search
        produces instances of the model by picking points in an N dimensional space.

        Parameters
        ----------
        analysis : af.Analysis
            An object that encapsulates the data and a log likelihood function.
        model : ModelMapper
            An object that represents possible instances of some model with a
            given dimensionality which is the number of free dimensions of the
            model.
        info : dict
            Optional dictionary containing information about the fit that can be loaded by the aggregator.

        Returns
        -------
        An object encapsulating how well the model fit the data, the best fit instance
        and an updated model with free parameters updated to represent beliefs
        produced by this fit.
        """

        self.paths.restore()

        if not os.path.exists(self.paths.has_completed_path) or not self.skip_completed:

            self.setup_log_file()
            self.save_model_info(model=model)
            self.save_parameter_names_file(model=model)
            self.save_metadata()
            self.save_info(info=info)
            self.save_search()
            self.save_model(model=model)

            self._fit(model=model, analysis=analysis)
            open(self.paths.has_completed_path, "w+").close()

            samples = self.perform_update(
                model=model, analysis=analysis, during_analysis=False
            )

        else:

            samples = self.samples_from_model(model=model)
            self.paths.backup_zip_remove()

        return Result(samples=samples, previous_model=model)

    @abstractmethod
    def _fit(self, model, analysis):
        pass

    @property
    def tag(self):
        """Tag the output folder of the non-linear search, based on the non linear search settings"""
        raise NotImplementedError

    def copy_with_name_extension(self, extension, remove_phase_tag=False):
        name = "{}/{}".format(self.paths.name, extension)

        if remove_phase_tag:
            tag = ""
        else:
            tag = self.paths.tag

        new_instance = self.__class__(
            paths=Paths(
                name=name,
                tag=tag,
                folders=self.paths.folders,
                path_prefix=self.paths.path_prefix,
                non_linear_name=self.paths.non_linear_name,
                remove_files=self.paths.remove_files,
            )
        )

        return new_instance

    @property
    def config_type(self):
        raise NotImplementedError()

    def config(self, section, attribute_name, attribute_type=str):
        """
        Get a config field from this search's section in non_linear.ini by a key and value type.

        Parameters
        ----------
        attribute_name: str
            The analysis_path of the field
        attribute_type: type
            The type of the value

        Returns
        -------
        attribute
            An attribute for the key with the specified type.
        """
        return self.config_type.config_for(self.__class__.__name__).get(
            section, attribute_name, attribute_type
        )

    def perform_update(self, model, analysis, during_analysis):
        """Perform an update of the non-linear search results, which occurs every *iterations_per_update* of the
        non-linear search. The update performs the following tasks:

        1) Visualize the maximum log likelihood model.
        2) Backup the samples.
        3) Output the model results to the model.reults file.

        These task are performed every n updates, set by the relevent *task_every_update* variable, for example
        *visualize_every_update* and *backup_every_update*.

        Parameters
        ----------
        model : ModelMapper
            The model which generates instances for different points in parameter space.
        analysis : Analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the non-linear search maximizes.
        during_analysis : bool
            If the update is during a non-linear search, in which case tasks are only performed after a certain number
             of updates and only a subset of visualization may be performed.
        """

        if self.should_backup() or not during_analysis:
            self.paths.backup()

        samples = self.samples_from_model(model=model)

        samples.write_table(filename=f"{self.paths.sym_path}/samples.csv")

        try:
            instance = samples.max_log_likelihood_instance
        except exc.FitException:
            return samples

        if self.should_visualize() or not during_analysis:
            analysis.visualize(instance=instance, during_analysis=during_analysis)

        if self.should_output_model_results() or not during_analysis:

            samples_text.results_to_file(
                samples=samples,
                file_results=self.paths.file_results,
                during_analysis=during_analysis,
            )

        if not during_analysis:
            self.paths.backup_zip_remove()

        return samples

    def setup_log_file(self):

        if not len(self.log_file) == 0:
            log_path = "{}/{}".format(self.paths.output_path, self.log_file)
            logger.handlers = [logging.FileHandler(log_path)]
            logger.propagate = False
            # noinspection PyProtectedMember
            logger.level = logging._nameToLevel[
                conf.instance.general.get("output", "log_level", str)
                .replace(" ", "")
                .upper()
            ]

    def save_model_info(self, model):
        """Save the model.info file, which summarizes every parameter and prior."""
        with open(self.paths.file_model_info, "w+") as f:
            f.write(model.info)

    def save_parameter_names_file(self, model):
        """Create the param_names file listing every parameter's label and Latex tag, which is used for *GetDist*
        visualization.

        The parameter labels are determined using the label.ini and label_format.ini config files."""

        paramnames_names = model.parameter_names
        paramnames_labels = model_text.parameter_labels_from_model(model=model)

        parameter_name_and_label = []

        for i in range(model.prior_count):
            line = formatter.label_and_label_string(
                label0=paramnames_names[i], label1=paramnames_labels[i], whitespace=70
            )
            parameter_name_and_label += [f"{line}\n"]

        formatter.output_list_of_strings_to_file(
            file=self.paths.file_param_names, list_of_strings=parameter_name_and_label
        )

    def save_info(self, info):
        """
        Save the dataset associated with the phase
        """
        with open("{}/info.pickle".format(self.paths.pickle_path), "wb") as f:
            pickle.dump(info, f)

    def save_search(self):
        """
        Save the seawrch associated with the phase as a pickle
        """
        with open(self.paths.make_non_linear_pickle_path(), "w+b") as f:
            f.write(pickle.dumps(self))

    def save_model(self, model):
        """
        Save the model associated with the phase as a pickle
        """
        with open(self.paths.make_model_pickle_path(), "w+b") as f:
            f.write(pickle.dumps(model))

    def save_metadata(self):
        """
        Save metadata associated with the phase, such as the name of the pipeline, the
        name of the phase and the name of the dataset being fit
        """
        with open("{}/metadata".format(self.paths.make_path()), "a") as f:
            f.write(self.make_metadata_text())

    @property
    def _default_metadata(self) -> Dict[str, str]:
        """
        A dictionary of metadata describing this phase, including the pipeline
        that it's embedded in.
        """
        return {
            "name": self.paths.name,
            "tag": self.paths.tag,
            "non_linear_search": type(self).__name__.lower(),
        }

    def make_metadata_text(self):
        return "\n".join(
            f"{key}={value or ''}" for key, value in {**self._default_metadata}.items()
        )

    def samples_from_model(self, model):
        raise NotImplementedError()

    def make_pool(self):
        """Make the pool instance used to parallelize a non-linear search alongside a set of unique ids for every
        process in the pool. If the specified number of cores is 1, a pool instance is not made and None is returned.

        The pool cannot be set as an attribute of the class itself because this prevents pickling, thus it is generated
        via this function before calling the non-linear search.

        The pool instance is also set up with a list of unique pool ids, which are used during model-fitting to
        identify a 'master core' (the one whose id value is lowest) which handles model result output, visualization,
        etc."""

        if self.number_of_cores == 1:

            return None, None

        else:

            manager = mp.Manager()
            idQueue = manager.Queue()

            [idQueue.put(i) for i in range(self.number_of_cores)]

            pool = mp.Pool(
                processes=self.number_of_cores, initializer=init, initargs=(idQueue,)
            )
            ids = pool.map(f, range(self.number_of_cores))

            return pool, [id[1] for id in ids]

    def __eq__(self, other):
        return isinstance(other, NonLinearSearch) and self.__dict__ == other.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.paths.restore()


class Analysis:
    def __init__(self, log_likelihood_cap=None):
        self.log_likelihood_cap = log_likelihood_cap

    def log_likelihood_function(self, instance):
        raise NotImplementedError()

    def visualize(self, instance, during_analysis):
        pass


class Result:
    """
    @DynamicAttrs
    """

    def __init__(self, samples, previous_model=None):
        """
        The result of an optimization.

        Parameters
        ----------
            A value indicating the figure of merit given by the optimal fit
        previous_model
            The model mapper from the stage that produced this result
        """

        self.samples = samples

        self.previous_model = previous_model

        self.__model = None

        self._instance = (
            samples.max_log_likelihood_instance if samples is not None else None
        )

    @property
    def log_likelihood(self):
        return max(self.samples.log_likelihoods)

    @property
    def instance(self):
        return self._instance

    @property
    def max_log_likelihood_instance(self):
        return self._instance

    @property
    def model(self):
        if self.__model is None:
            self.__model = self.previous_model.mapper_from_gaussian_tuples(
                self.samples.gaussian_priors_at_sigma(sigma=3.0)
            )
        return self.__model

    @model.setter
    def model(self, model):
        self.__model = model

    def __str__(self):
        return "Analysis Result:\n{}".format(
            "\n".join(
                ["{}: {}".format(key, value) for key, value in self.__dict__.items()]
            )
        )

    def model_absolute(self, a: float) -> mm.ModelMapper:
        """
        Parameters
        ----------
        a
            The absolute width of gaussian priors

        Returns
        -------
        A model mapper created by taking results from this phase and creating priors with the defined absolute
        width.
        """
        return self.previous_model.mapper_from_gaussian_tuples(
            self.samples.gaussian_priors_at_sigma(sigma=3.0), a=a
        )

    def model_relative(self, r: float) -> mm.ModelMapper:
        """
        Parameters
        ----------
        r
            The relative width of gaussian priors

        Returns
        -------
        A model mapper created by taking results from this phase and creating priors with the defined relative
        width.
        """
        return self.previous_model.mapper_from_gaussian_tuples(
            self.samples.gaussian_priors_at_sigma(sigma=3.0), r=r
        )


class IntervalCounter:
    def __init__(self, interval):
        self.count = 0
        self.interval = interval

    def __call__(self):
        if self.interval == -1:
            return False
        self.count += 1
        return self.count % self.interval == 0


def init(queue):
    global idx
    idx = queue.get()


def f(x):
    global idx
    process = mp.current_process()
    sleep(1)
    return (idx, process.pid, x * x)



from autoconf import conf

from autofit.non_linear.abstract_search import NonLinearSearch


class AbstractOptimizer(NonLinearSearch):
    @property
    def config_type(self):
        return conf.instance.optimize



import math

from autoconf import conf
from autofit import exc
from autofit.non_linear.samples import PDFSamples
from autofit.non_linear.abstract_search import NonLinearSearch
from autofit.non_linear.abstract_search import Analysis
from autofit.non_linear.abstract_search import Result


class MockSearch(NonLinearSearch):
    def _fit(self, model, analysis):

        if model.prior_count == 0:
            raise AssertionError("There are no priors associated with the model!")
        if model.prior_count != len(model.unique_prior_paths):
            raise AssertionError(
                "Prior count doesn't match number of unique prior paths"
            )
        index = 0
        unit_vector = model.prior_count * [0.5]
        while True:
            try:
                instance = model.instance_from_unit_vector(unit_vector)
                fit = analysis.log_likelihood_function(instance)
                break
            except exc.FitException as e:
                unit_vector[index] += 0.1
                if unit_vector[index] >= 1:
                    raise e
                index = (index + 1) % model.prior_count
        return Result(
            previous_model=model,
            samples=MockSamples(
                log_likelihoods=fit,
                model=model,
                gaussian_tuples=[
                    (prior.mean, prior.width if math.isfinite(prior.width) else 1.0)
                    for prior in sorted(model.priors, key=lambda prior: prior.id)
                ],
            ),
        )

    @property
    def config_type(self):
        return conf.instance.mock

    @property
    def tag(self):
        return "mock"

    def samples_from_model(self, model):
        return MockOutput()

    @property
    def name(self):
        return "mock_nlo"


class MockOutput(object):
    def __init__(self):
        pass


class MockAnalysis(Analysis):
    def log_likelihood_function(self, instance):
        return 1.0

    def visualize(self, instance, during_analysis):
        pass

    def __init__(self, data):
        self.data = data


class MockSamples(PDFSamples):
    def __init__(
        self,
        model=None,
        max_log_likelihood_instance=None,
        log_likelihoods=None,
        gaussian_tuples=None,
    ):

        super().__init__(
            model=model, parameters=[], log_likelihoods=[], log_priors=[], weights=[]
        )

        self._max_log_likelihood_instance = max_log_likelihood_instance
        self.log_likelihoods = log_likelihoods
        self.gaussian_tuples = gaussian_tuples

    @property
    def max_log_likelihood_instance(self) -> int:
        return self._max_log_likelihood_instance

    def gaussian_priors_at_sigma(self, sigma=None):
        return self.gaussian_tuples



import math

import autofit as af
from test_autofit import mock


def test_constructor():
    prior_model = af.PriorModel(mock.Circle)

    assert prior_model.prior_count == 1

    instance = prior_model.instance_from_prior_medians()

    assert instance.radius == 1.0
    assert instance.circumference == 2 * math.pi


def test_alternative():
    prior_model = af.PriorModel(mock.Circle.with_circumference)

    assert prior_model.prior_count == 1

    instance = prior_model.instance_from_prior_medians()

    assert instance.circumference == 1.0
    assert instance.radius == 1.0 / (2 * math.pi)



import logging

from autofit import exc

logger = logging.getLogger(__name__)


class ResultsCollection:
    def __init__(self):
        """
        A collection of results from previous phases. Results can be obtained using an index or the name of the phase
        from whence they came.
        """
        self.__result_list = []
        self.__result_dict = {}

    def copy(self):
        collection = ResultsCollection()
        collection.__result_dict = self.__result_dict
        collection.__result_list = self.__result_list
        return collection

    @property
    def reversed(self):
        return reversed(self.__result_list)

    @property
    def last(self):
        """
        The result of the last phase
        """
        if len(self.__result_list) > 0:
            return self.__result_list[-1]
        return None

    @property
    def first(self):
        """
        The result of the first phase
        """
        if len(self.__result_list) > 0:
            return self.__result_list[0]
        return None

    def add(self, phase_name, result):
        """
        Add the result of a phase.

        Parameters
        ----------
        phase_name: str
            The name of the phase
        result
            The result of that phase
        """
        try:
            self.__result_list[self.__result_list.index(result)] = result
        except ValueError:
            self.__result_list.append(result)
        self.__result_dict[phase_name] = result

    def __getitem__(self, item):
        """
        Get the result of a previous phase by index

        Parameters
        ----------
        item: int
            The index of the result

        Returns
        -------
        result: Result
            The result of a previous phase
        """
        return self.__result_list[item]

    def __len__(self):
        return len(self.__result_list)

    def from_phase(self, phase_name):
        """
        Returns the result of a previous phase by its name

        Parameters
        ----------
        phase_name: str
            The name of a previous phase

        Returns
        -------
        result: Result
            The result of that phase

        Raises
        ------
        exc.PipelineException
            If no phase with the expected result is found
        """
        try:
            return self.__result_dict[phase_name]
        except KeyError:
            raise exc.PipelineException(
                "No previous phase named {} found in results ({})".format(
                    phase_name, ", ".join(self.__result_dict.keys())
                )
            )

    def __contains__(self, item):
        return item in self.__result_dict


class Pipeline:
    def __init__(self, pipeline_name, *phases):
        """
        A pipeline of phases to be run sequentially. Results are passed between phases. Phases must have unique names.

        Parameters
        ----------
        pipeline_name: str
            The phase_name of this pipeline
        """
        self.pipeline_name = pipeline_name
        self.phases = phases
        self.pipeline_tag = None

        for phase in phases:
            if phase.pipeline_name is None:
                phase.pipeline_name = pipeline_name
            if phase.pipeline_tag is None:
                phase.pipeline_tag = self.pipeline_tag

            with open(phase.paths.file_model_promises, "w+") as f:
                f.write(phase.model.info)

        phase_names = [phase.phase_name for phase in phases]

        if len(set(phase_names)) < len(phase_names):
            raise exc.PipelineException(
                "Cannot create pipelines with duplicate phase names. ({})".format(
                    ", ".join(phase_names)
                )
            )

    def __getitem__(self, item):
        return self.phases[item]

    def __add__(self, other):
        """
        Compose two runners

        Parameters
        ----------
        other: Pipeline
            Another pipeline

        Returns
        -------
        composed_pipeline: Pipeline
            A pipeline that runs all the  phases from this pipeline and then all the phases from the other pipeline
        """
        return self.__class__(
            "{} + {}".format(
                self.pipeline_name,
                other.pipeline_name
            ),
            *(self.phases + other.phases)
        )

    def run(self, dataset):
        def runner(phase, results):
            return phase.run(dataset=dataset, results=results)

        return self.run_function(runner)

    def run_function(self, func):
        """
        Run the function for each phase in the pipeline.

        Parameters
        ----------
        func
            A function that takes a phase and prior results, returning results for that phase

        Returns
        -------
        results: ResultsCollection
            A collection of results
        """
        results = ResultsCollection()
        for i, phase in enumerate(self.phases):
            logger.info(
                "Running Phase {} (Number {})".format(
                    phase.phase_name,
                    i
                )
            )
            name = phase.phase_name
            results.add(name, func(phase, results))
        return results



import os
from astropy.io import fits
from test_autofit.integration.src.model import profiles

import numpy as np

# %%
#%matplotlib inline

# %%
"""
___Simulator___
This script simulates the 1D Gaussians line profile datasets used by the integration tests.
"""


def numpy_array_1d_to_fits(array_1d, file_path, overwrite=False):
    """Write a 1D NumPy array to a .fits file.

    Parameters
    ----------
    array_1d : ndarray
        The 1D array that is written to fits.
    file_path : str
        The full path of the file that is output, including the file name and '.fits' extension.
    overwrite : bool
        If True and a file already exists with the input file_path the .fits file is overwritten. If False, an error \
        will be raised.

    Returns
    -------
    None

    Examples
    --------
    array_1d = np.ones(shape=(,5))
    numpy_array_1d_to_fits(array_1d=array_1d, file_path='/path/to/file/filename.fits', overwrite=True)
    """

    if overwrite and os.path.exists(file_path):
        os.remove(file_path)
        
    new_hdr = fits.Header()
    hdu = fits.PrimaryHDU(array_1d, new_hdr)
    hdu.writeto(file_path)


# %%
"""
The path to the chapter and dataset folder on your computer. The data should be distributed with PyAutoFit, however
if you wish to reuse this script to generate it again (or new datasets) you must update the paths appropriately.
"""

# %%
test_path = "{}".format(os.path.dirname(os.path.realpath(__file__)))
dataset_path = f"{test_path}/dataset"

# %%
"""
__Gaussian X1__

Setup the path and filename the .fits file of the Gaussian is written to.
"""

# %%
data_path = f"{dataset_path}/gaussian_x1"

# %%
"""
Create a model instance of the Gaussian.
"""

# %%
gaussian = profiles.Gaussian(centre=50.0, intensity=25.0, sigma=10.0)

# %%
"""
Specify the number of pixels used to create the xvalues on which the 1D line of the profile is generated using and
thus defining the number of data-points in our data.
"""

# %%
pixels = 100
xvalues = np.arange(pixels)

# %%
"""
Evaluate this Gaussian model instance at every xvalues to create its model line profile.
"""

# %%
model_line = gaussian.line_from_xvalues(xvalues=xvalues)

# %%
"""
Determine the noise (at a specified signal to noise level) in every pixel of our model line profile.
"""

# %%
signal_to_noise_ratio = 25.0
noise = np.random.normal(0.0, 1.0 / signal_to_noise_ratio, pixels)

# %%
"""
Add this noise to the model line to create the line data that is fitted, using the signal-to-noise ratio to compute
noise-map of our data which is required when evaluating the chi-squared value of the likelihood.
"""

# %%
data = model_line + noise
noise_map = (1.0 / signal_to_noise_ratio) * np.ones(pixels)

# %%
"""
Output this data to fits file, so it can be loaded and fitted in the HowToFit tutorials.
"""

# %%
numpy_array_1d_to_fits(
    array_1d=data, file_path=f"{data_path}/data.fits", overwrite=True
)
numpy_array_1d_to_fits(
    array_1d=noise_map, file_path=f"{data_path}/noise_map.fits", overwrite=True
)


import copy

import pytest

import autofit as af
from test_autofit import mock


@pytest.fixture(name="instance_prior_model")
def make_instance_prior_model():
    instance = mock.SimpleClass(1.0, 2.0)
    return af.AbstractPriorModel.from_instance(instance)


@pytest.fixture(name="list_prior_model")
def make_list_prior_model():
    instance = [mock.SimpleClass(1.0, 2.0)]
    return af.AbstractPriorModel.from_instance(instance)


@pytest.fixture(name="complex_prior_model")
def make_complex_prior_model():
    instance = mock.ComplexClass(mock.SimpleClass(1.0, 2.0))
    return af.AbstractPriorModel.from_instance(instance)


class TestAsModel:
    def test_instance(self, instance_prior_model):
        model = instance_prior_model.as_model()
        assert model.prior_count == 2

    def test_list(self, list_prior_model):
        model = list_prior_model.as_model()
        assert isinstance(model, af.CollectionPriorModel)
        assert model.prior_count == 2

    def test_complex(self, complex_prior_model):
        assert complex_prior_model.prior_count == 0
        model = complex_prior_model.as_model()
        assert model.prior_count == 2
        assert model.simple.prior_count == 2

    def test_galaxy_list(self):
        galaxies = af.ModelInstance()
        galaxies.one = mock.Galaxy()
        instance = af.ModelInstance()
        instance.galaxies = galaxies
        model = instance.as_model(model_classes=(mock.Galaxy,))
        assert model.prior_count == 1


class TestFromInstance:
    def test_model_mapper(self):
        instance = af.ModelInstance()
        instance.simple = mock.SimpleClass(1.0, 2.0)

        result = af.AbstractPriorModel.from_instance(instance)

        assert isinstance(result, af.ModelMapper)

    def test_with_model_classes(self):
        instance = mock.ComplexClass(mock.SimpleClass(1.0, 2.0))
        model = af.AbstractPriorModel.from_instance(
            instance, model_classes=(mock.SimpleClass,)
        )
        assert model.prior_count == 2

    def test_list_with_model_classes(self):
        instance = [
            mock.SimpleClass(1.0, 2.0),
            mock.ComplexClass(mock.SimpleClass(1.0, 2.0)),
        ]
        model = af.AbstractPriorModel.from_instance(
            instance, model_classes=(mock.ComplexClass,)
        )

        assert model.prior_count == 2
        assert model[0].prior_count == 0
        assert model[1].prior_count == 2

    def test_dict_with_model_classes(self):
        instance = {
            "one": mock.SimpleClass(1.0, 2.0),
            "two": mock.ComplexClass(mock.SimpleClass(1.0, 2.0)),
        }
        model = af.AbstractPriorModel.from_instance(
            instance, model_classes=(mock.ComplexClass,)
        )

        assert model.prior_count == 2
        assert model[0].prior_count == 0
        assert model[1].prior_count == 2

        assert model.one.one == 1.0
        assert model.one.two == 2.0

        assert isinstance(model.two.simple.one, af.Prior)
        assert isinstance(model.two.simple.two, af.Prior)

    def test_instance(self, instance_prior_model):
        assert instance_prior_model.cls == mock.SimpleClass
        assert instance_prior_model.prior_count == 0
        assert instance_prior_model.one == 1.0
        assert instance_prior_model.two == 2.0

        new_instance = instance_prior_model.instance_for_arguments({})
        assert isinstance(new_instance, mock.SimpleClass)
        assert new_instance.one == 1.0
        assert new_instance.two == 2.0

    def test_complex(self, complex_prior_model):
        assert complex_prior_model.cls == mock.ComplexClass
        assert complex_prior_model.prior_count == 0
        assert isinstance(complex_prior_model.simple, af.PriorModel)
        assert complex_prior_model.simple.cls == mock.SimpleClass
        assert complex_prior_model.simple.one == 1.0

        new_instance = complex_prior_model.instance_for_arguments({})
        assert isinstance(new_instance, mock.ComplexClass)
        assert isinstance(new_instance.simple, mock.SimpleClass)
        assert new_instance.simple.one == 1.0

    def test_list(self, list_prior_model):
        assert isinstance(list_prior_model, af.CollectionPriorModel)
        assert isinstance(list_prior_model[0], af.PriorModel)
        assert list_prior_model[0].one == 1.0

    def test_dict(self):
        instance = {"simple": mock.SimpleClass(1.0, 2.0)}
        prior_model = af.AbstractPriorModel.from_instance(instance)
        assert isinstance(prior_model, af.CollectionPriorModel)
        assert isinstance(prior_model.simple, af.PriorModel)
        assert prior_model.simple.one == 1.0

        new_instance = prior_model.instance_for_arguments({})
        assert isinstance(new_instance.simple, mock.SimpleClass)

        prior_model = af.AbstractPriorModel.from_instance(new_instance)
        assert isinstance(prior_model, af.CollectionPriorModel)
        assert isinstance(prior_model.simple, af.PriorModel)
        assert prior_model.simple.one == 1.0

    def test_dimension_types(self):
        instance = mock.MockDistanceClass(mock.MockDistance(1.0), mock.MockDistance(2.0))
        result = af.AbstractPriorModel.from_instance(
            instance, model_classes=(mock.MockDistanceClass,)
        )
        assert isinstance(result.first, af.PriorModel)

        new_instance = result.instance_from_unit_vector([0.1, 0.2])
        assert isinstance(new_instance, mock.MockDistanceClass)
        assert isinstance(new_instance.first, mock.MockDistance)
        assert new_instance.first == 0.1


class TestSum:
    def test_add_prior_models(self):
        profile_1 = af.PriorModel(mock.EllipticalProfile)
        profile_2 = af.PriorModel(mock.EllipticalProfile)

        profile_1.axis_ratio = 1.0
        profile_2.phi = 0.0

        result = profile_1 + profile_2

        assert isinstance(result, af.PriorModel)
        assert result.cls == mock.EllipticalProfile
        assert isinstance(result.axis_ratio, af.Prior)
        assert isinstance(result.phi, af.Prior)

    def test_fail_for_mismatch(self):
        profile_1 = af.PriorModel(mock.EllipticalProfile)
        profile_2 = af.PriorModel(mock.EllipticalMassProfile)

        with pytest.raises(TypeError):
            profile_1 + profile_2

    def test_add_children(self):
        galaxy_1 = af.PriorModel(
            mock.Galaxy,
            light_profiles=af.CollectionPriorModel(light_1=mock.EllipticalProfile),
            mass_profiles=af.CollectionPriorModel(mass_1=mock.EllipticalMassProfile),
        )
        galaxy_2 = af.PriorModel(
            mock.Galaxy,
            light_profiles=af.CollectionPriorModel(light_2=mock.EllipticalProfile),
            mass_profiles=af.CollectionPriorModel(mass_2=mock.EllipticalMassProfile),
        )

        result = galaxy_1 + galaxy_2

        assert result.light_profiles.light_1 == galaxy_1.light_profiles.light_1
        assert result.light_profiles.light_2 == galaxy_2.light_profiles.light_2

        assert result.mass_profiles.mass_1 == galaxy_1.mass_profiles.mass_1
        assert result.mass_profiles.mass_2 == galaxy_2.mass_profiles.mass_2

    def test_prior_model_override(self):
        galaxy_1 = af.PriorModel(
            mock.Galaxy,
            light_profiles=af.CollectionPriorModel(light=mock.EllipticalProfile()),
            mass_profiles=af.CollectionPriorModel(mass=mock.EllipticalMassProfile),
        )
        galaxy_2 = af.PriorModel(
            mock.Galaxy,
            light_profiles=af.CollectionPriorModel(light=mock.EllipticalProfile),
            mass_profiles=af.CollectionPriorModel(mass=mock.EllipticalMassProfile()),
        )

        result = galaxy_1 + galaxy_2

        assert result.mass_profiles.mass == galaxy_1.mass_profiles.mass
        assert result.light_profiles.light == galaxy_2.light_profiles.light


class TestFloatAnnotation:
    def test_distance_from_distance(self):
        original = mock.MockDistance(1.0)
        # noinspection PyTypeChecker
        distance = mock.MockDistanceClass(first=original, second=2.0)

        assert distance.first is original

    # noinspection PyTypeChecker
    def test_instantiate_distance(self):
        distance = mock.MockDistanceClass(first=1.0, second=2.0)

        assert distance.first == 1.0
        assert distance.second == 2.0

        assert isinstance(distance.first, mock.MockDistance)
        assert isinstance(distance.second, mock.MockDistance)

        distance = mock.MockDistanceClass(1.0, 2.0)

        assert distance.first == 1.0
        assert distance.second == 2.0

        assert isinstance(distance.first, mock.MockDistance)
        assert isinstance(distance.second, mock.MockDistance)

    def test_distance(self):
        mapper = af.ModelMapper()
        mapper.object = mock.MockDistanceClass

        assert mapper.prior_count == 2

        result = mapper.instance_from_unit_vector([0.5, 1.0])
        assert isinstance(result.object, mock.MockDistanceClass)
        assert result.object.first == 0.5
        assert result.object.second == 1.0

        assert isinstance(result.object.first, mock.MockDistance)

    def test_position(self):
        mapper = af.ModelMapper()
        mapper.object = mock.PositionClass

        assert mapper.prior_count == 2
        result = mapper.instance_from_unit_vector([0.5, 1.0])
        assert isinstance(result.object, mock.PositionClass)
        assert result.object.position[0] == 0.5
        assert result.object.position[1] == 1.0

        assert isinstance(result.object.position[0], mock.MockDistance)
        assert isinstance(result.object.position[1], mock.MockDistance)

    # noinspection PyUnresolvedReferences
    def test_prior_linking(self):
        mapper = af.ModelMapper()
        mapper.a = mock.SimpleClass
        mapper.b = mock.SimpleClass

        assert mapper.prior_count == 4

        mapper.a.one = mapper.b.one

        assert mapper.prior_count == 3

        mapper.a.two = mapper.b.two

        assert mapper.prior_count == 2

        mapper.a.one = mapper.a.two
        mapper.b.one = mapper.b.two

        assert mapper.prior_count == 1

    def test_prior_tuples(self):
        prior_model = af.PriorModel(mock.MockDistanceClass)

        assert prior_model.unique_prior_tuples[0].name == "first"
        assert prior_model.unique_prior_tuples[1].name == "second"


class TestHashing:
    def test_is_hashable(self):
        assert hash(af.AbstractPriorModel()) is not None
        assert hash(af.PriorModel(mock.SimpleClass)) is not None
        assert (
            hash(af.AnnotationPriorModel(mock.SimpleClass, mock.SimpleClass, "one"))
            is not None
        )

    def test_prior_prior_model_hash_consecutive(self):
        prior = af.UniformPrior(0, 1)
        prior_model = af.AbstractPriorModel()

        assert prior.id + 1 == prior_model.id


class StringDefault:
    def __init__(self, value="a string"):
        self.value = value


class TestStringArguments:
    def test_string_default(self):
        prior_model = af.PriorModel(StringDefault)
        assert prior_model.prior_count == 0

        assert prior_model.instance_for_arguments({}).value == "a string"


class TestPriorModelArguments:
    def test_list_arguments(self):
        prior_model = af.PriorModel(mock.ListClass)

        assert prior_model.prior_count == 0

        prior_model = af.PriorModel(mock.ListClass, ls=[mock.SimpleClass])

        assert prior_model.prior_count == 2

        prior_model = af.PriorModel(
            mock.ListClass, ls=[mock.SimpleClass, mock.SimpleClass]
        )

        assert prior_model.prior_count == 4

    def test_float_argument(self):
        prior = af.UniformPrior(0.5, 2.0)
        prior_model = af.PriorModel(mock.Galaxy, redshift=prior)

        assert prior_model.prior_count == 1
        assert prior_model.priors[0] is prior

        prior_model = af.PriorModel(mock.Galaxy, redshift=4.0)
        assert prior_model.prior_count == 0
        assert prior_model.redshift == 4.0

        instance = prior_model.instance_for_arguments({})
        assert instance.redshift == 4.0

    def test_model_argument(self):
        lens_galaxy = af.PriorModel(mock.Galaxy)
        source_galaxy = mock.Galaxy()
        tracer = af.PriorModel(
            mock.Tracer,
            lens_galaxy=lens_galaxy,
            source_galaxy=source_galaxy
        )

        assert tracer.lens_galaxy is lens_galaxy
        assert tracer.prior_count == 1

        deferred_instance = tracer.instance_for_arguments({lens_galaxy.redshift: 0.5})
        instance = deferred_instance(grid=None)

        assert instance.lens_galaxy.redshift == 0.5
        assert instance.source_galaxy is source_galaxy

    def test_no_passing(self):
        mapper = af.ModelMapper()
        mapper.distance = mock.MockDistanceClass
        instance = mapper.instance_from_prior_medians()
        assert not hasattr(instance.distance.first, "value") or not isinstance(
            instance.distance.first.value, af.Prior
        )

    def test_arbitrary_keyword_arguments(self):
        prior_model = af.PriorModel(
            mock.Galaxy,
            light=mock.EllipticalCoredIsothermal,
            mass=mock.EllipticalMassProfile,
        )
        assert prior_model.prior_count == 11
        instance = prior_model.instance_from_unit_vector(
            [0.5] * prior_model.prior_count
        )
        assert isinstance(instance.light, mock.EllipticalCoredIsothermal)
        assert isinstance(instance.mass, mock.EllipticalMassProfile)


class TestCase:
    def test_complex_class(self):
        prior_model = af.PriorModel(mock.ComplexClass)

        assert hasattr(prior_model, "simple")
        assert prior_model.simple.prior_count == 2
        assert prior_model.prior_count == 2

    def test_create_instance(self):
        mapper = af.ModelMapper()
        mapper.complex = mock.ComplexClass

        instance = mapper.instance_from_unit_vector([1.0, 0.0])

        assert instance.complex.simple.one == 1.0
        assert instance.complex.simple.two == 0.0

    def test_instantiate_with_list_arguments(self):
        mapper = af.ModelMapper()
        mapper.list_object = af.PriorModel(
            mock.ListClass, ls=[mock.SimpleClass, mock.SimpleClass]
        )

        assert len(mapper.list_object.ls) == 2

        assert mapper.list_object.prior_count == 4
        assert mapper.prior_count == 4

        instance = mapper.instance_from_unit_vector([0.1, 0.2, 0.3, 0.4])

        assert len(instance.list_object.ls) == 2
        assert instance.list_object.ls[0].one == 0.1
        assert instance.list_object.ls[0].two == 0.2
        assert instance.list_object.ls[1].one == 0.3
        assert instance.list_object.ls[1].two == 0.4

    def test_mix_instances_and_models(self):
        mapper = af.ModelMapper()
        mapper.list_object = af.PriorModel(
            mock.ListClass, ls=[mock.SimpleClass, mock.SimpleClass(1, 2)]
        )

        assert mapper.prior_count == 2

        instance = mapper.instance_from_unit_vector([0.1, 0.2])

        assert len(instance.list_object.ls) == 2
        assert instance.list_object.ls[0].one == 0.1
        assert instance.list_object.ls[0].two == 0.2
        assert instance.list_object.ls[1].one == 1
        assert instance.list_object.ls[1].two == 2


class TestCollectionPriorModel:
    def test_keyword_arguments(self):
        prior_model = af.CollectionPriorModel(
            one=mock.SimpleClass, two=mock.SimpleClass(1, 2)
        )

        assert len(prior_model.direct_prior_model_tuples) == 1
        assert len(prior_model) == 2

        instance = prior_model.instance_for_arguments(
            {prior_model.one.one: 0.1, prior_model.one.two: 0.2}
        )

        assert instance.one.one == 0.1
        assert instance.one.two == 0.2

        assert instance.two.one == 1
        assert instance.two.two == 2

    def test_mix_instances_in_list_prior_model(self):
        prior_model = af.CollectionPriorModel(
            [mock.SimpleClass, mock.SimpleClass(1, 2)]
        )

        assert len(prior_model.direct_prior_model_tuples) == 1
        assert prior_model.prior_count == 2

        mapper = af.ModelMapper()
        mapper.ls = prior_model

        instance = mapper.instance_from_unit_vector([0.1, 0.2])

        assert len(instance.ls) == 2

        assert instance.ls[0].one == 0.1
        assert instance.ls[0].two == 0.2
        assert instance.ls[1].one == 1
        assert instance.ls[1].two == 2

        assert len(prior_model.prior_class_dict) == 2

    def test_list_in_list_prior_model(self):
        prior_model = af.CollectionPriorModel([[mock.SimpleClass]])

        assert len(prior_model.direct_prior_model_tuples) == 1
        assert prior_model.prior_count == 2

    def test_list_prior_model_with_dictionary(self, simple_model):
        assert isinstance(simple_model.simple, af.PriorModel)

    def test_override_with_instance(self, simple_model):
        simple_instance = mock.SimpleClass(1, 2)

        simple_model.simple = simple_instance

        assert len(simple_model) == 1
        assert simple_model.simple == simple_instance

    def test_names_of_priors(self):
        collection = af.CollectionPriorModel([af.UniformPrior(), af.UniformPrior()])
        assert collection.name_for_prior(collection[0]) == "0"


@pytest.fixture(name="simple_model")
def make_simple_model():
    return af.CollectionPriorModel({"simple": mock.SimpleClass})


class TestCopy:
    def test_simple(self, simple_model):
        assert simple_model.prior_count > 0
        assert copy.deepcopy(simple_model).prior_count == simple_model.prior_count

    def test_embedded(self, simple_model):
        model = af.CollectionPriorModel(simple=simple_model)
        assert copy.deepcopy(model).prior_count == model.prior_count

    def test_circular(self):
        one = af.PriorModel(mock.SimpleClass)

        one.one = af.PriorModel(mock.SimpleClass)
        one.one.one = one

        # noinspection PyUnresolvedReferences
        assert one.prior_count == one.one.prior_count
        assert copy.deepcopy(one).prior_count == one.prior_count



import pytest

import autofit as af


@pytest.fixture(
    name="prior"
)
def make_prior():
    return af.UniformPrior()


class TestAddition:
    def test_prior_plus_prior(self, prior):
        sum_prior = prior + prior
        assert sum_prior.instance_from_unit_vector([1.0]) == 2.0

    def test_negative_prior(self, prior):
        negative = -prior
        assert negative.instance_from_unit_vector([1.0]) == -1.0

    def test_prior_minus_prior(self, prior):
        sum_prior = prior - prior
        assert sum_prior.instance_from_unit_vector([1.0]) == 0.0

    def test_prior_plus_float(self, prior):
        sum_prior = prior + 1.0
        assert sum_prior.instance_from_unit_vector([1.0]) == 2.0

    def test_float_plus_prior(self, prior):
        sum_prior = 1.0 + prior
        assert sum_prior.instance_from_unit_vector([1.0]) == 2.0


class TestMultiplication:
    def test_prior_times_prior(self, prior):
        multiple_prior = (prior + prior) * (prior + prior)
        assert multiple_prior.instance_from_unit_vector([1.0]) == 4

    def test_prior_times_float(self, prior):
        multiple_prior = prior * 2.0
        assert multiple_prior.instance_from_unit_vector([1.0]) == 2.0

    def test_float_times_prior(self, prior):
        multiple_prior = 2.0 * prior
        assert multiple_prior.instance_from_unit_vector([1.0]) == 2.0


class TestDivision:
    def test_prior_over_prior(self, prior):
        division_prior = prior / prior
        assert division_prior.instance_from_unit_vector(
            [10.0],
            assert_priors_in_limits=False
        ) == 1

    def test_prior_over_float(self, prior):
        division_prior = prior / 2
        assert division_prior.instance_from_unit_vector(
            [1.0]
        ) == 0.5

    def test_float_over_prior(self, prior):
        division_prior = 4.0 / prior
        assert division_prior.instance_from_unit_vector(
            [0.5]
        ) == 8.0


class TestFloorDiv:
    def test_prior_over_int(self, prior):
        division_prior = prior // 2
        assert division_prior.instance_from_unit_vector(
            [3.0],
            assert_priors_in_limits=False
        ) == 1.0

    def test_int_over_prior(self, prior):
        division_prior = 3 // prior
        assert division_prior.instance_from_unit_vector(
            [2.0],
            assert_priors_in_limits=False
        ) == 1.0


class TestMod:
    def test_prior_mod_int(self, prior):
        mod_prior = prior % 3
        assert mod_prior.instance_from_unit_vector(
            [5.0],
            assert_priors_in_limits=False
        ) == 2.0

    def test_int_mod_prior(self, prior):
        mod_prior = 5.0 % prior
        assert mod_prior.instance_from_unit_vector(
            [3.0],
            assert_priors_in_limits=False
        ) == 2.0


def test_abs(prior):
    prior = af.UniformPrior(
        -1, 0
    )
    assert prior.value_for(0.0) == -1
    prior = abs(
        prior
    )
    assert prior.instance_from_unit_vector(
        [0.0]
    ) == 1.0


class TestPowers:
    def test_prior_to_prior(self, prior):
        power_prior = prior ** prior
        assert power_prior.instance_from_unit_vector(
            [2.0],
            assert_priors_in_limits=False
        ) == 4.0

    def test_prior_to_float(self, prior):
        power_prior = prior ** 3
        assert power_prior.instance_from_unit_vector(
            [2.0],
            assert_priors_in_limits=False
        ) == 8.0

    def test_float_to_prior(self, prior):
        power_prior = 3.0 ** prior
        assert power_prior.instance_from_unit_vector(
            [2.0],
            assert_priors_in_limits=False
        ) == 9.0


class TestInequality:
    def test_prior_lt_prior(self, prior):
        inequality_prior = (prior * prior) < prior
        assert inequality_prior.instance_from_unit_vector(
            [0.5],
            assert_priors_in_limits=False
        ) is True
        inequality_prior = (prior * prior) > prior
        assert inequality_prior.instance_from_unit_vector(
            [0.5],
            assert_priors_in_limits=False
        ) is False



from . import conf
from . import exc
from .aggregator import Aggregator
from .aggregator import PhaseOutput
from .mapper import link
from .mapper import prior
from .mapper.model import AbstractModel
from .mapper.model import ModelInstance
from .mapper.model import ModelInstance as Instance
from .mapper.model import path_instances_of_class
from .mapper.model_mapper import ModelMapper
from .mapper.model_mapper import ModelMapper as Mapper
from .mapper.model_object import ModelObject
from .mapper.prior import AbstractPromise
from .mapper.prior import GaussianPrior
from .mapper.prior import LogUniformPrior
from .mapper.prior import Prior
from .mapper.prior import UniformPrior
from .mapper.prior import last
from .mapper.prior.deferred import DeferredArgument
from .mapper.prior.deferred import DeferredInstance
from .mapper.prior_model.abstract import AbstractPriorModel
from .mapper.prior_model.annotation import AnnotationPriorModel
from .mapper.prior_model.attribute_pair import AttributeNameValue
from .mapper.prior_model.attribute_pair import InstanceNameValue
from .mapper.prior_model.attribute_pair import PriorNameValue
from .mapper.prior_model.attribute_pair import cast_collection
from .mapper.prior_model.collection import CollectionPriorModel
from .mapper.prior_model.collection import CollectionPriorModel as Collection
from .mapper.prior_model.dimension_type import DimensionType
from .mapper.prior_model.dimension_type import map_types
from .mapper.prior_model.prior_model import PriorModel
from .mapper.prior_model.prior_model import PriorModel as Model
from .mapper.prior_model.util import PriorModelNameValue
from .non_linear.abstract_search import Analysis
from .non_linear.abstract_search import NonLinearSearch
from .non_linear.abstract_search import Result
from .non_linear.optimize.downhill_simplex import DownhillSimplex
from .non_linear.optimize.pyswarms import PySwarmsLocal
from .non_linear.optimize.pyswarms import PySwarmsGlobal
from .non_linear.mcmc.emcee import Emcee
from .non_linear.mock.mock_nlo import MockSearch
from .non_linear.nest.dynesty import DynestyDynamic
from .non_linear.nest.dynesty import DynestyStatic
from .non_linear.nest.multi_nest import MultiNest
from .non_linear.initializer import InitializerPrior
from .non_linear.initializer import InitializerBall
from .non_linear.paths import Paths
from .non_linear.paths import convert_paths
from .non_linear.paths import make_path
from .non_linear.samples import OptimizerSamples
from .non_linear.samples import PDFSamples
from .non_linear.samples import MCMCSamples
from .non_linear.samples import NestSamples
from .non_linear.grid_search import GridSearch as NonLinearSearchGridSearch
from .non_linear.grid_search import GridSearchResult
from .text import Model
from .text import formatter
from .text import samples_text
from .tools import util
from .tools.phase import AbstractPhase
from .tools.phase import Dataset
from .tools.phase import Phase
from .tools.phase import as_grid_search
from .tools.phase import AbstractPhaseSettings
from .tools.phase_property import PhaseProperty
from .tools.pipeline import Pipeline
from .tools.pipeline import ResultsCollection

__version__ = '0.60.1'



import re
import shutil
from os import walk
from uuid import uuid1


class Line:
    def __init__(self, string):
        if "*" in string:
            print("Please ensure no imports in the __init__ contain a *")
            exit(1)
        if "," in string:
            print("Comma separated imports not allowed")
            exit(1)
        self.string = string.replace("\n", "")
        self.id = str(uuid1())

    @property
    def sources(self):
        return (
            self.source,
            self.joint_source
        )

    @property
    def joint_source(self):
        return f"{self.import_target[1:]}.{self.source}"

    def __str__(self):
        return f"{self.source} -> {self.import_target}"

    def __repr__(self):
        return f"<{self.__class__.__name__} {self}>"

    def __len__(self):
        return len(self.source)

    def __lt__(self, other):
        return len(self) < len(other)

    def __gt__(self, other):
        return len(self) > len(other)

    def __eq__(self, other):
        return self._target == other._target

    @property
    def is_import(self):
        return self.string.startswith("from")

    @property
    def source(self):
        match = re.match("from .* as (.+)", self.string)
        if match is not None:
            return match.group(1)
        return re.match("from .* import (.+)", self.string).group(1)

    @property
    def target(self):
        return self._target.split(".")[-1]

    @property
    def import_target(self):
        target = ".".join(
            self._target.split(".")[:-1]
        )
        if len(target) > 0:
            return f".{target}"
        return target

    def __hash__(self):
        return hash(self._target)

    @property
    def _target(self):
        return self.string.replace(
            f" as {self.source}",
            ""
        ).replace(
            "from ",
            ""
        ).replace(
            " import ",
            "."
        ).lstrip(
            "."
        )


class Converter:
    def __init__(self, name, prefix, lines):
        self.name = name
        self.prefix = prefix
        self.lines = sorted(
            filter(
                lambda line: line.is_import,
                lines
            ),
            reverse=True
        )

    @classmethod
    def from_prefix_and_source_directory(
            cls,
            name,
            prefix,
            source_directory
    ):
        source_directory = source_directory
        with open(
                f"{source_directory}/__init__.py"
        ) as f:
            lines = map(Line, f.readlines())
        return Converter(name, prefix, lines)

    def convert(self, string):
        matched_lines = set()
        string = string.replace(f"import {self.name} as {self.prefix}", "")
        for line in self.lines:
            for line_source in line.sources:
                source = f"{self.prefix}.{line_source}"
                if source in string:
                    matched_lines.add(
                        line
                    )
                    string = string.replace(
                        source,
                        line.id
                    )
        for line in self.lines:
            string = string.replace(
                line.id,
                line.target
            )

        import_string = "\n".join(
            f"from {self.name}{line.import_target} import {line.target}"
            for line in matched_lines
        )
        return f"{import_string}\n{string}"


def edenise(
        root_directory,
        name,
        prefix
):
    target_directory = f"{root_directory}/../eden/{name}_eden"

    print(f"Creating {target_directory}...")
    shutil.copytree(
        root_directory,
        target_directory,
        symlinks=True
    )

    converter = Converter.from_prefix_and_source_directory(
        name=name,
        prefix=prefix,
        source_directory=f"{root_directory}/{name}"
    )

    for root, _, files in walk(f"{target_directory}/test_{name}"):
        for file in files:
            if file.endswith(".py"):
                with open(f"{root}/{file}", "r+") as f:
                    string = f.read()
                    f.seek(0)
                    f.write(
                        converter.convert(
                            string
                        )
                    )
                    f.truncate()

    open(f"{target_directory}/{name}/__init__.py", "w+").close()



import numpy as np
import scipy.optimize

from autofit import exc
from autofit.non_linear.optimize.abstract_optimize import AbstractOptimizer
from autofit.non_linear.abstract_search import logger
from autofit.non_linear.paths import Paths


class DownhillSimplex(AbstractOptimizer):
    def __init__(self, paths=None, fmin=scipy.optimize.fmin):

        if paths is None:
            paths = Paths()

        super().__init__(paths)

        self.xtol = self.config("search", "xtol", float)
        self.ftol = self.config("search", "ftol", float)
        self.maxiter = self.config("search", "maxiter", int)
        self.maxfun = self.config("search", "maxfun", int)

        self.full_output = self.config("search", "full_output", int)
        self.disp = self.config("search", "disp", int)
        self.retall = self.config("search", "retall", int)

        self.fmin = fmin

        logger.debug("Creating DownhillSimplex NLO")

    @property
    def tag(self):
        return "dhs"

    def copy_with_name_extension(self, extension, remove_phase_tag=False):
        copy = super().copy_with_name_extension(
            extension=extension, remove_phase_tag=remove_phase_tag
        )
        copy.fmin = self.fmin
        copy.xtol = self.xtol
        copy.ftol = self.ftol
        copy.maxiter = self.maxiter
        copy.maxfun = self.maxfun
        copy.full_output = self.full_output
        copy.disp = self.disp
        copy.retall = self.retall
        return copy

    class Fitness(AbstractOptimizer.Fitness):
        def __init__(self, paths, model, analysis, samples_fom_model):

            super().__init__(
                paths=paths,
                analysis=analysis,
                model=model,
                samples_from_model=samples_fom_model,
            )

        def __call__(self, vector):
            try:
                instance = self.model.instance_from_vector(vector)
                log_likelihood = self.fit_instance(instance)
            except exc.FitException:
                log_likelihood = -np.inf
            return -2 * log_likelihood

    def _fit(self, model, analysis):

        initial_vector = model.physical_values_from_prior_medians

        fitness_function = DownhillSimplex.Fitness(
            paths=self.paths,
            model=model,
            analysis=analysis,
            samples_fom_model=self.samples_from_model,
        )

        logger.info("Running DownhillSimplex...")
        samples = self.fmin(fitness_function, x0=initial_vector)
        logger.info("DownhillSimplex complete")

        res = fitness_function.result

        # Create a set of Gaussian priors from this result and associate them with the result object.
        res.gaussian_tuples = [(mean, 0) for mean in samples]
        res.previous_model = model

        analysis.visualize(instance=res.instance, during_analysis=False)
        self.paths.backup_zip_remove()
        return res

    def samples_from_model(self, model):
        pass




def create_path(path, folders):
    """ For a given path, create a directory structure composed of a set of folders and return the path to the \
    inner-most folder.

    For example, if path='/path/to/folders', and folder_names=['folder1', 'folder2'], the directory created will be
    '/path/to/folders/folder1/folder2/' and the returned path will be '/path/to/folders/folder1/folder2/'.

    If the folders already exist, routine continues as normal.

    Parameters
    ----------
    path : str
        The path where the directories are created.
    folders : [str]
        The names of the folders which are created in the path directory.

    Returns
    -------
    path
        A string specifying the path to the inner-most folder created.

    Examples
    --------
    path = '/path/to/folders'
    path = make_and_return_path(path=path, folder_names=['folder1', 'folder2'].
    """
    for folder in folders:

        path = os.path.join(path, folder)

        try:
            os.makedirs(path)
        except FileExistsError:
            pass

    return f"{path}/"


from contextlib import contextmanager
import sys, os

@contextmanager
def suppress_stdout():
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout


import os

import pytest

import autofit as af
from test_autofit.mock import GeometryProfile, Galaxy


@pytest.fixture(name="results")
def make_results_collection():
    results = af.ResultsCollection()

    results.add("first phase", "one")
    results.add("second phase", "two")

    return results


class TestResultsCollection:
    def test_with_name(self, results):
        assert results.from_phase("first phase") == "one"
        assert results.from_phase("second phase") == "two"

    def test_with_index(self, results):
        assert results[0] == "one"
        assert results[1] == "two"
        assert results.first == "one"
        assert results.last == "two"
        assert len(results) == 2

    def test_missing_result(self, results):
        with pytest.raises(af.exc.PipelineException):
            results.from_phase("third phase")


class MockPhase(af.AbstractPhase):
    def make_result(self, result, analysis):
        pass

    @af.convert_paths
    def __init__(self, paths, search):
        super().__init__(paths=paths, search=search)

    def save_metadata(self, *args, **kwargs):
        pass


class TestPipeline:
    def test_unique_phases(self):

        phase_1 = MockPhase(paths=af.Paths(name="one"), search=af.MockSearch())
        phase_2 = MockPhase(paths=af.Paths(name="two"), search=af.MockSearch())

        af.Pipeline("name", phase_1, phase_2)
        with pytest.raises(af.exc.PipelineException):
            af.Pipeline("name", MockPhase(af.Paths("one")), MockPhase(af.Paths("one")))

    def test_search_assertion(self, model):
        paths = af.Paths("Phase Name")
        search = af.MockSearch(paths)
        phase = MockPhase(phase_name="Phase_Name", search=search)
        phase.model.profile = GeometryProfile

        try:
            os.makedirs(phase.paths.make_path())
        except FileExistsError:
            pass



        phase.model.profile.centre_0 = af.UniformPrior()

    def test_name_composition(self):
        first = af.Pipeline("first")
        second = af.Pipeline("second")

        assert (first + second).pipeline_name == "first + second"



# noinspection PyUnresolvedReferences
class TestPhasePipelineName:
    def test_name_stamping(self):
        one = MockPhase(af.Paths("one"))
        two = MockPhase(af.Paths("two"))
        af.Pipeline("name", one, two)

        assert one.pipeline_name == "name"
        assert two.pipeline_name == "name"

    def test_no_restamping(self):
        one = MockPhase(af.Paths("one"))
        two = MockPhase(af.Paths("two"))
        pipeline_one = af.Pipeline("one", one)
        pipeline_two = af.Pipeline("two", two)

        composed_pipeline = pipeline_one + pipeline_two

        assert composed_pipeline[0].pipeline_name == "one"
        assert composed_pipeline[1].pipeline_name == "two"

        assert one.pipeline_name == "one"
        assert two.pipeline_name == "two"



import os
from os import path

import pytest

import autofit as af

directory = path.dirname(path.realpath(__file__))


def test_dynamic(phase):
    paths = phase.paths
    non_linear_tag = paths.non_linear_tag

    phase.search.n_live_points += 1

    assert paths.non_linear_tag != non_linear_tag


class PatchPaths(af.Paths):
    @property
    def path(self):
        return f"{directory}/path"

    @property
    def backup_path(self) -> str:
        return f"{directory}/backup_path"

    @property
    def sym_path(self) -> str:
        return f"{directory}/sym_path"

    @property
    @af.make_path
    def output_path(self) -> str:
        return f"{directory}/phase_output_path"


@pytest.fixture(name="paths")
def make_paths():
    paths = PatchPaths(remove_files=True)
    return paths


def test_backup_zip_remove(paths):
    os.mkdir(paths.sym_path)
    os.mkdir(paths.path)

    paths.backup_zip_remove()

    assert not os.path.exists(paths.path)
    assert not os.path.exists(f"{directory}/phase_output_path")
    assert os.path.exists(paths.backup_path)
    assert os.path.exists(paths.zip_path)

    os.remove(paths.zip_path)
    os.rmdir(paths.sym_path)
    os.rmdir(paths.backup_path)


def test_restore(paths):
    os.mkdir(paths.sym_path)
    os.mkdir(paths.path)

    paths.backup_zip_remove()

    os.rmdir(paths.sym_path)

    paths.restore()

    assert os.path.exists(paths.output_path)
    assert os.path.exists(paths.backup_path)
    assert not os.path.exists(paths.zip_path)

    os.rmdir(paths.backup_path)
    os.rmdir(paths.output_path)



import pytest

import autofit as af
from test_autofit import mock
from test_autofit.mock import MockClassx2, MockProfile


@pytest.fixture(name="galaxy_1")
def make_galaxy_1():
    return mock.Galaxy()


@pytest.fixture(name="galaxy_2")
def make_galaxy_2():
    return mock.Galaxy()


@pytest.fixture(name="instance")
def make_instance(galaxy_1, galaxy_2):
    sub = af.ModelInstance()

    instance = af.ModelInstance()
    sub.galaxy_1 = galaxy_1

    instance.galaxy_2 = galaxy_2
    instance.sub = sub

    sub_2 = af.ModelInstance()
    sub_2.galaxy_1 = galaxy_1

    instance.sub.sub = sub_2

    return instance


class TestModelInstance:
    def test_as_model(self, instance):

        model = instance.as_model()
        assert isinstance(model, af.ModelMapper)
        assert isinstance(model.galaxy_2, af.PriorModel)
        assert model.galaxy_2.cls == mock.Galaxy

    def test_object_for_path(self, instance, galaxy_1, galaxy_2):

        assert instance.object_for_path(("galaxy_2",)) is galaxy_2
        assert instance.object_for_path(("sub", "galaxy_1")) is galaxy_1
        assert instance.object_for_path(("sub", "sub", "galaxy_1")) is galaxy_1
        setattr(instance.object_for_path(("galaxy_2",)), "galaxy", galaxy_1)
        assert galaxy_2.galaxy is galaxy_1

    def test_path_instance_tuples_for_class(self, instance, galaxy_1, galaxy_2):
        result = instance.path_instance_tuples_for_class(mock.Galaxy)
        assert result[0] == (("galaxy_2",), galaxy_2)
        assert result[1] == (("sub", "galaxy_1"), galaxy_1)
        assert result[2] == (("sub", "sub", "galaxy_1"), galaxy_1)

    def test_simple_model(self):
        mapper = af.ModelMapper()

        mapper.mock_class = MockClassx2

        model_map = mapper.instance_from_unit_vector([1.0, 1.0])

        assert isinstance(model_map.mock_class, MockClassx2)
        assert model_map.mock_class.one == 1.0
        assert model_map.mock_class.two == 1.0

    def test_two_object_model(self):
        mapper = af.ModelMapper()

        mapper.mock_class_1 = MockClassx2
        mapper.mock_class_2 = MockClassx2

        model_map = mapper.instance_from_unit_vector([1.0, 0.0, 0.0, 1.0])

        assert isinstance(model_map.mock_class_1, MockClassx2)
        assert isinstance(model_map.mock_class_2, MockClassx2)

        assert model_map.mock_class_1.one == 1.0
        assert model_map.mock_class_1.two == 0.0

        assert model_map.mock_class_2.one == 0.0
        assert model_map.mock_class_2.two == 1.0

    def test_swapped_prior_construction(self):
        mapper = af.ModelMapper()

        mapper.mock_class_1 = MockClassx2
        mapper.mock_class_2 = MockClassx2

        # noinspection PyUnresolvedReferences
        mapper.mock_class_2.one = mapper.mock_class_1.one

        model_map = mapper.instance_from_unit_vector([1.0, 0.0, 0.0])

        assert isinstance(model_map.mock_class_1, MockClassx2)
        assert isinstance(model_map.mock_class_2, MockClassx2)

        assert model_map.mock_class_1.one == 1.0
        assert model_map.mock_class_1.two == 0.0

        assert model_map.mock_class_2.one == 1.0
        assert model_map.mock_class_2.two == 0.0

    def test_prior_replacement(self):
        mapper = af.ModelMapper()

        mapper.mock_class = MockClassx2

        mapper.mock_class.one = af.UniformPrior(100, 200)

        model_map = mapper.instance_from_unit_vector([0.0, 0.0])

        assert model_map.mock_class.one == 100.0

    def test_tuple_arg(self):
        mapper = af.ModelMapper()

        mapper.mock_profile = MockProfile

        model_map = mapper.instance_from_unit_vector([1.0, 0.0, 0.0])

        assert model_map.mock_profile.centre == (1.0, 0.0)
        assert model_map.mock_profile.intensity == 0.0

    def test_modify_tuple(self):
        mapper = af.ModelMapper()

        mapper.mock_profile = MockProfile

        # noinspection PyUnresolvedReferences
        mapper.mock_profile.centre.centre_0 = af.UniformPrior(1.0, 10.0)

        model_map = mapper.instance_from_unit_vector([1.0, 1.0, 1.0])

        assert model_map.mock_profile.centre == (10.0, 1.0)

    def test_match_tuple(self):
        mapper = af.ModelMapper()

        mapper.mock_profile = MockProfile

        # noinspection PyUnresolvedReferences
        mapper.mock_profile.centre.centre_1 = mapper.mock_profile.centre.centre_0

        model_map = mapper.instance_from_unit_vector([1.0, 0.0])

        assert model_map.mock_profile.centre == (1.0, 1.0)
        assert model_map.mock_profile.intensity == 0.0



from abc import ABC, abstractmethod
from typing import List, Iterator

from .phase_output import PhaseOutput


class AttributePredicate:
    def __init__(self, *path):
        """
        Used to produce predicate objects for filtering in the aggregator.

        When an unrecognised attribute is called on an aggregator an instance
        of this object is created. This object implements comparison methods
        facilitating construction of predicates.

        Parameters
        ----------
        path
            A series of names of attributes that can be used to get a value.
            For example, (mask, pixel_size) would get the pixel size of a mask
            when evaluated for a given phase.
        """
        self.path = path

    def value_for_phase(
            self,
            phase: PhaseOutput
    ):
        """
        Recurse the phase output by iterating the attributes in the path
        and getting a value for each attribute.
        """
        value = phase
        for attribute in self.path:
            value = getattr(
                value,
                attribute
            )
        return value

    def __eq__(self, value):
        """
        Create a predicate which asks whether the given value is equal to
        the attribute of a phase.
        """
        return EqualityPredicate(
            self,
            value
        )

    def __le__(self, other):
        return OrPredicate(
            self == other,
            self < other
        )

    def __ge__(self, other):
        return OrPredicate(
            self == other,
            self > other
        )

    def __getattr__(self, item: str) -> "AttributePredicate":
        """
        Adds another item to the path
        """
        return AttributePredicate(
            *self.path, item
        )

    def __gt__(self, other):
        """
        Is the value of this attribute for a given phase greater than some
        other value?
        """
        return GreaterThanPredicate(
            self, other
        )

    def __lt__(self, other):
        """
        Is the value of this attribute for a given phase less than some
        other value?
        """
        return LessThanPredicate(
            self, other
        )

    def __ne__(self, other):
        """
        Create a predicate which asks whether the given value is not equal to
        the attribute of a phase.
        """
        return ~(self == other)

    def contains(self, value):
        """
        Create a predicate which asks whether the given is contained within
        the attribute of a phase.
        """
        return ContainsPredicate(
            self,
            value
        )


class AbstractPredicate(ABC):
    """
    Comparison between a value and some attribute of a phase
    """

    def filter(
            self,
            phases: List[PhaseOutput]
    ) -> Iterator[PhaseOutput]:
        """
        Only return phases for which this predicate evaluates to True

        Parameters
        ----------
        phases

        Returns
        -------

        """
        return filter(
            lambda phase: self(phase),
            phases
        )

    def __invert__(self) -> "NotPredicate":
        """
        A predicate that evaluates to True when this predicate evaluates
        to False
        """
        return NotPredicate(
            self
        )

    def __or__(self, other: "AbstractPredicate") -> "OrPredicate":
        """
        Returns a predicate that is true if either predicate is true
        for a given phase.
        """
        return OrPredicate(self, other)

    def __and__(self, other: "AbstractPredicate") -> "AndPredicate":
        """
        Returns a predicate that is true if both predicates are true
        for a given phase.
        """
        return AndPredicate(self, other)

    @abstractmethod
    def __call__(self, phase: PhaseOutput) -> bool:
        """
        Does the attribute of the phase match the requirement of this predicate?
        """


class CombinationPredicate(AbstractPredicate, ABC):
    def __init__(
            self,
            one: AbstractPredicate,
            two: AbstractPredicate
    ):
        """
        Abstract predicate combining two other predicates.

        Parameters
        ----------
        one
        two
            Child predicates
        """
        self.one = one
        self.two = two


class OrPredicate(CombinationPredicate):
    def __call__(self, phase: PhaseOutput):
        """
        The disjunction of two predicates.

        Parameters
        ----------
        phase
            An object representing the output of a given phase.

        Returns
        -------
        True if either predicate is True for the phase
        """
        return self.one(phase) or self.two(phase)


class AndPredicate(CombinationPredicate):
    def __call__(self, phase: PhaseOutput):
        """
        The conjunction of two predicates.

        Parameters
        ----------
        phase
            An object representing the output of a given phase.

        Returns
        -------
        True if both predicates are True for the phase
        """
        return self.one(phase) and self.two(phase)


class ComparisonPredicate(AbstractPredicate, ABC):
    def __init__(
            self,
            attribute_predicate: AttributePredicate,
            value
    ):
        """
        Compare an attribute of a phase with a value.

        Parameters
        ----------
        attribute_predicate
            An attribute path of a phase
        value
            A value to which the attribute is compared
        """
        self.attribute_predicate = attribute_predicate
        self._value = value

    def value(
            self,
            phase
    ):
        if isinstance(self._value, AttributePredicate):
            return self._value.value_for_phase(
                phase
            )
        return self._value


class GreaterThanPredicate(ComparisonPredicate):
    def __call__(
            self,
            phase: PhaseOutput
    ) -> bool:
        """
        Parameters
        ----------
        phase
            An object representing the output of a given phase.

        Returns
        -------
        True iff the value of the attribute of the phase is greater than
        the value associated with this predicate
        """

        return self.attribute_predicate.value_for_phase(
            phase
        ) > self.value(
            phase
        )


class LessThanPredicate(ComparisonPredicate):
    def __call__(
            self,
            phase: PhaseOutput
    ) -> bool:
        """
        Parameters
        ----------
        phase
            An object representing the output of a given phase.

        Returns
        -------
        True iff the value of the attribute of the phase is less than
        the value associated with this predicate
        """
        return self.attribute_predicate.value_for_phase(
            phase
        ) < self.value(
            phase
        )


class ContainsPredicate(ComparisonPredicate):
    def __call__(
            self,
            phase: PhaseOutput
    ) -> bool:
        """
        Parameters
        ----------
        phase
            An object representing the output of a given phase.

        Returns
        -------
        True iff the value of the attribute of the phase contains
        the value associated with this predicate
        """
        return self.value(
            phase
        ) in self.attribute_predicate.value_for_phase(
            phase
        )


class EqualityPredicate(ComparisonPredicate):
    def __call__(self, phase):
        """
        Parameters
        ----------
        phase
            An object representing the output of a given phase.

        Returns
        -------
        True iff the value of the attribute of the phase is equal to
        the value associated with this predicate
        """
        try:
            value = self.value(
                phase
            )
        except AttributeError:
            value = self.value
        return self.attribute_predicate.value_for_phase(
            phase
        ) == value


class NotPredicate(AbstractPredicate):
    def __init__(
            self,
            predicate: AbstractPredicate
    ):
        """
        Negates the output of a predicate.

        If the predicate would have returned True for a given phase
        it now returns False and vice-versa.

        Parameters
        ----------
        predicate
            A predicate that is negated
        """
        self.predicate = predicate

    def __call__(self, phase: PhaseOutput) -> bool:
        """
        Evaluate the predicate for the phase and return the negation
        of the result.

        Parameters
        ----------
        phase
            The output of an AutoFit phase

        Returns
        -------
        The negation of the underlying predicate
        """
        return not self.predicate(
            phase
        )
